import json


class Error(Exception):
    def __init__(self, code: str, message: str):
        self.code = code
        self.message = message

    def serialize(self) -> str:
        err = {"code": self.code, "message": self.message}
        return json.dumps(err)


RuntimeNotInitializedError = Error("", "Runtime was not initialized")



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: plugin.proto
# Protobuf Python Version: 5.26.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import struct_pb2 as google_dot_protobuf_dot_struct__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0cplugin.proto\x12\x05proto\x1a\x1cgoogle/protobuf/struct.proto\"2\n\x05\x45rror\x12\n\n\x02OK\x18\x02 \x01(\x08\x12\x0c\n\x04\x63ode\x18\x03 \x01(\t\x12\x0f\n\x07message\x18\x04 \x01(\t\"1\n\x0bInitRequest\x12\x14\n\x0c\x65vent_server\x18\x01 \x01(\r\x12\x0c\n\x04port\x18\x02 \x01(\r\"/\n\x0fOnCreateRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0e\n\x06\x63onfig\x18\x02 \x01(\x0c\"/\n\x10OnCreateResponse\x12\x1b\n\x05\x65rror\x18\x01 \x01(\x0b\x32\x0c.proto.Error\"3\n\x10OnMessageRequest\x12\x0c\n\x04guid\x18\x01 \x01(\t\x12\x11\n\tinMessage\x18\x02 \x01(\x0c\"D\n\x11OnMessageResponse\x12\x12\n\noutMessage\x18\x01 \x01(\x0c\x12\x1b\n\x05\x65rror\x18\x02 \x01(\x0b\x32\x0c.proto.Error\"\x1e\n\x0eOnCloseRequest\x12\x0c\n\x04guid\x18\x01 \x01(\t\".\n\x0fOnCloseResponse\x12\x1b\n\x05\x65rror\x18\x01 \x01(\x0b\x32\x0c.proto.Error\"0\n\x18PGetCapabilitiesResponse\x12\x14\n\x0c\x63\x61pabilities\x18\x01 \x01(\x04\"\x07\n\x05\x45mpty\"&\n\x11IsRunningResponse\x12\x11\n\tisRunning\x18\x01 \x01(\x08\";\n\x0c\x44\x65\x62ugRequest\x12\x0c\n\x04guid\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x0f\n\x07message\x18\x03 \x01(\x0c\"2\n\x14\x45mitFlowEventRequest\x12\x0c\n\x04guid\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\"/\n\x10\x45mitInputRequest\x12\x0c\n\x04guid\x18\x01 \x01(\t\x12\r\n\x05input\x18\x02 \x01(\x0c\"?\n\x11\x45mitOutputRequest\x12\x0c\n\x04guid\x18\x01 \x01(\t\x12\x0e\n\x06output\x18\x02 \x01(\x0c\x12\x0c\n\x04port\x18\x03 \x01(\x05\"?\n\x10\x45mitErrorRequest\x12\x0c\n\x04guid\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x0f\n\x07message\x18\x03 \x01(\t\"6\n\x13GetVaultItemRequest\x12\x0f\n\x07vaultId\x18\x01 \x01(\t\x12\x0e\n\x06ItemId\x18\x02 \x01(\t\"=\n\x14GetVaultItemResponse\x12%\n\x04item\x18\x01 \x01(\x0b\x32\x17.google.protobuf.Struct\"D\n\x13SetVaultItemRequest\x12\x0f\n\x07vaultId\x18\x01 \x01(\t\x12\x0e\n\x06ItemId\x18\x02 \x01(\t\x12\x0c\n\x04\x44\x61ta\x18\x03 \x01(\x0c\"=\n\x14SetVaultItemResponse\x12%\n\x04item\x18\x01 \x01(\x0b\x32\x17.google.protobuf.Struct\"8\n\x08Variable\x12\r\n\x05scope\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x0f\n\x07payload\x18\x03 \x01(\x0c\"7\n\x12GetVariableRequest\x12!\n\x08variable\x18\x01 \x01(\x0b\x32\x0f.proto.Variable\"=\n\x13GetVariableResponse\x12&\n\x05value\x18\x01 \x01(\x0b\x32\x17.google.protobuf.Struct\"_\n\x12SetVariableRequest\x12!\n\x08variable\x18\x01 \x01(\x0b\x32\x0f.proto.Variable\x12&\n\x05value\x18\x02 \x01(\x0b\x32\x17.google.protobuf.Struct\">\n\x14GetRobotInfoResponse\x12&\n\x05robot\x18\x01 \x01(\x0b\x32\x17.google.protobuf.Struct\"5\n\x11\x41ppRequestRequest\x12\x0f\n\x07request\x18\x01 \x01(\x0c\x12\x0f\n\x07timeout\x18\x02 \x01(\x05\"&\n\x13\x41ppRequestV2Request\x12\x0f\n\x07request\x18\x01 \x01(\x0c\"&\n\x12\x41ppRequestResponse\x12\x10\n\x08response\x18\x01 \x01(\x0c\"$\n\x11\x41ppPublishRequest\x12\x0f\n\x07request\x18\x01 \x01(\x0c\"0\n\x13\x44ownloadFileRequest\x12\x0b\n\x03url\x18\x01 \x01(\t\x12\x0c\n\x04path\x18\x02 \x01(\t\"A\n\x12\x41ppDownloadRequest\x12\x11\n\tdirectory\x18\x01 \x01(\t\x12\x0c\n\x04\x66ile\x18\x02 \x01(\t\x12\n\n\x02id\x18\x03 \x01(\t\"#\n\x13\x41ppDownloadResponse\x12\x0c\n\x04path\x18\x01 \x01(\t\",\n\x10\x41ppUploadRequest\x12\n\n\x02id\x18\x01 \x01(\t\x12\x0c\n\x04path\x18\x02 \x01(\t\" \n\x11\x41ppUploadResponse\x12\x0b\n\x03url\x18\x01 \x01(\t\"\xb3\x01\n\x15GatewayRequestRequest\x12\x0e\n\x06method\x18\x01 \x01(\t\x12\x10\n\x08\x65ndpoint\x18\x02 \x01(\t\x12:\n\x07headers\x18\x03 \x03(\x0b\x32).proto.GatewayRequestRequest.HeadersEntry\x12\x0c\n\x04\x62ody\x18\x04 \x01(\t\x1a.\n\x0cHeadersEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"\xa7\x01\n\x16GatewayRequestResponse\x12\x12\n\nstatusCode\x18\x01 \x01(\x05\x12;\n\x07headers\x18\x02 \x03(\x0b\x32*.proto.GatewayRequestResponse.HeadersEntry\x12\x0c\n\x04\x62ody\x18\x03 \x01(\t\x1a.\n\x0cHeadersEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"\x9a\x01\n\x0bHttpRequest\x12\x0e\n\x06method\x18\x01 \x01(\t\x12\x0b\n\x03url\x18\x02 \x01(\t\x12\x30\n\x07headers\x18\x03 \x03(\x0b\x32\x1f.proto.HttpRequest.HeadersEntry\x12\x0c\n\x04\x62ody\x18\x04 \x01(\x0c\x1a.\n\x0cHeadersEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"\x94\x01\n\x0cHttpResponse\x12\x13\n\x0bstatus_code\x18\x01 \x01(\x05\x12\x31\n\x07headers\x18\x02 \x03(\x0b\x32 .proto.HttpResponse.HeadersEntry\x12\x0c\n\x04\x62ody\x18\x03 \x01(\x0c\x1a.\n\x0cHeadersEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"9\n\x08NodeInfo\x12\x0c\n\x04type\x18\x01 \x01(\t\x12\x0f\n\x07version\x18\x02 \x01(\t\x12\x0e\n\x06\x63onfig\x18\x03 \x01(\x0c\"7\n\x19GetPortConnectionsRequest\x12\x0c\n\x04guid\x18\x01 \x01(\t\x12\x0c\n\x04port\x18\x02 \x01(\x05\"<\n\x1aGetPortConnectionsResponse\x12\x1e\n\x05nodes\x18\x01 \x03(\x0b\x32\x0f.proto.NodeInfo2\xa9\x02\n\x04Node\x12(\n\x04Init\x12\x12.proto.InitRequest\x1a\x0c.proto.Empty\x12;\n\x08OnCreate\x12\x16.proto.OnCreateRequest\x1a\x17.proto.OnCreateResponse\x12>\n\tOnMessage\x12\x17.proto.OnMessageRequest\x1a\x18.proto.OnMessageResponse\x12\x38\n\x07OnClose\x12\x15.proto.OnCloseRequest\x1a\x16.proto.OnCloseResponse\x12@\n\x0fGetCapabilities\x12\x0c.proto.Empty\x1a\x1f.proto.PGetCapabilitiesResponse2\x9d\n\n\rRuntimeHelper\x12#\n\x05\x43lose\x12\x0c.proto.Empty\x1a\x0c.proto.Empty\x12*\n\x05\x44\x65\x62ug\x12\x13.proto.DebugRequest\x1a\x0c.proto.Empty\x12:\n\rEmitFlowEvent\x12\x1b.proto.EmitFlowEventRequest\x1a\x0c.proto.Empty\x12\x32\n\tEmitInput\x12\x17.proto.EmitInputRequest\x1a\x0c.proto.Empty\x12\x34\n\nEmitOutput\x12\x18.proto.EmitOutputRequest\x1a\x0c.proto.Empty\x12\x32\n\tEmitError\x12\x17.proto.EmitErrorRequest\x1a\x0c.proto.Empty\x12G\n\x0cGetVaultItem\x12\x1a.proto.GetVaultItemRequest\x1a\x1b.proto.GetVaultItemResponse\x12G\n\x0cSetVaultItem\x12\x1a.proto.SetVaultItemRequest\x1a\x1b.proto.SetVaultItemResponse\x12\x44\n\x0bGetVariable\x12\x19.proto.GetVariableRequest\x1a\x1a.proto.GetVariableResponse\x12\x36\n\x0bSetVariable\x12\x19.proto.SetVariableRequest\x1a\x0c.proto.Empty\x12\x39\n\x0cGetRobotInfo\x12\x0c.proto.Empty\x1a\x1b.proto.GetRobotInfoResponse\x12\x41\n\nAppRequest\x12\x18.proto.AppRequestRequest\x1a\x19.proto.AppRequestResponse\x12\x45\n\x0c\x41ppRequestV2\x12\x1a.proto.AppRequestV2Request\x1a\x19.proto.AppRequestResponse\x12\x34\n\nAppPublish\x12\x18.proto.AppPublishRequest\x1a\x0c.proto.Empty\x12\x38\n\x0c\x44ownloadFile\x12\x1a.proto.DownloadFileRequest\x1a\x0c.proto.Empty\x12\x44\n\x0b\x41ppDownload\x12\x19.proto.AppDownloadRequest\x1a\x1a.proto.AppDownloadResponse\x12>\n\tAppUpload\x12\x17.proto.AppUploadRequest\x1a\x18.proto.AppUploadResponse\x12M\n\x0eGatewayRequest\x12\x1c.proto.GatewayRequestRequest\x1a\x1d.proto.GatewayRequestResponse\x12\x37\n\x0cProxyRequest\x12\x12.proto.HttpRequest\x1a\x13.proto.HttpResponse\x12Y\n\x12GetPortConnections\x12 .proto.GetPortConnectionsRequest\x1a!.proto.GetPortConnectionsResponse\x12\x33\n\tIsRunning\x12\x0c.proto.Empty\x1a\x18.proto.IsRunningResponseB$Z\"robomotion/robomotion-runner/protob\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'plugin_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'Z\"robomotion/robomotion-runner/proto'
  _globals['_GATEWAYREQUESTREQUEST_HEADERSENTRY']._loaded_options = None
  _globals['_GATEWAYREQUESTREQUEST_HEADERSENTRY']._serialized_options = b'8\001'
  _globals['_GATEWAYREQUESTRESPONSE_HEADERSENTRY']._loaded_options = None
  _globals['_GATEWAYREQUESTRESPONSE_HEADERSENTRY']._serialized_options = b'8\001'
  _globals['_HTTPREQUEST_HEADERSENTRY']._loaded_options = None
  _globals['_HTTPREQUEST_HEADERSENTRY']._serialized_options = b'8\001'
  _globals['_HTTPRESPONSE_HEADERSENTRY']._loaded_options = None
  _globals['_HTTPRESPONSE_HEADERSENTRY']._serialized_options = b'8\001'
  _globals['_ERROR']._serialized_start=53
  _globals['_ERROR']._serialized_end=103
  _globals['_INITREQUEST']._serialized_start=105
  _globals['_INITREQUEST']._serialized_end=154
  _globals['_ONCREATEREQUEST']._serialized_start=156
  _globals['_ONCREATEREQUEST']._serialized_end=203
  _globals['_ONCREATERESPONSE']._serialized_start=205
  _globals['_ONCREATERESPONSE']._serialized_end=252
  _globals['_ONMESSAGEREQUEST']._serialized_start=254
  _globals['_ONMESSAGEREQUEST']._serialized_end=305
  _globals['_ONMESSAGERESPONSE']._serialized_start=307
  _globals['_ONMESSAGERESPONSE']._serialized_end=375
  _globals['_ONCLOSEREQUEST']._serialized_start=377
  _globals['_ONCLOSEREQUEST']._serialized_end=407
  _globals['_ONCLOSERESPONSE']._serialized_start=409
  _globals['_ONCLOSERESPONSE']._serialized_end=455
  _globals['_PGETCAPABILITIESRESPONSE']._serialized_start=457
  _globals['_PGETCAPABILITIESRESPONSE']._serialized_end=505
  _globals['_EMPTY']._serialized_start=507
  _globals['_EMPTY']._serialized_end=514
  _globals['_ISRUNNINGRESPONSE']._serialized_start=516
  _globals['_ISRUNNINGRESPONSE']._serialized_end=554
  _globals['_DEBUGREQUEST']._serialized_start=556
  _globals['_DEBUGREQUEST']._serialized_end=615
  _globals['_EMITFLOWEVENTREQUEST']._serialized_start=617
  _globals['_EMITFLOWEVENTREQUEST']._serialized_end=667
  _globals['_EMITINPUTREQUEST']._serialized_start=669
  _globals['_EMITINPUTREQUEST']._serialized_end=716
  _globals['_EMITOUTPUTREQUEST']._serialized_start=718
  _globals['_EMITOUTPUTREQUEST']._serialized_end=781
  _globals['_EMITERRORREQUEST']._serialized_start=783
  _globals['_EMITERRORREQUEST']._serialized_end=846
  _globals['_GETVAULTITEMREQUEST']._serialized_start=848
  _globals['_GETVAULTITEMREQUEST']._serialized_end=902
  _globals['_GETVAULTITEMRESPONSE']._serialized_start=904
  _globals['_GETVAULTITEMRESPONSE']._serialized_end=965
  _globals['_SETVAULTITEMREQUEST']._serialized_start=967
  _globals['_SETVAULTITEMREQUEST']._serialized_end=1035
  _globals['_SETVAULTITEMRESPONSE']._serialized_start=1037
  _globals['_SETVAULTITEMRESPONSE']._serialized_end=1098
  _globals['_VARIABLE']._serialized_start=1100
  _globals['_VARIABLE']._serialized_end=1156
  _globals['_GETVARIABLEREQUEST']._serialized_start=1158
  _globals['_GETVARIABLEREQUEST']._serialized_end=1213
  _globals['_GETVARIABLERESPONSE']._serialized_start=1215
  _globals['_GETVARIABLERESPONSE']._serialized_end=1276
  _globals['_SETVARIABLEREQUEST']._serialized_start=1278
  _globals['_SETVARIABLEREQUEST']._serialized_end=1373
  _globals['_GETROBOTINFORESPONSE']._serialized_start=1375
  _globals['_GETROBOTINFORESPONSE']._serialized_end=1437
  _globals['_APPREQUESTREQUEST']._serialized_start=1439
  _globals['_APPREQUESTREQUEST']._serialized_end=1492
  _globals['_APPREQUESTV2REQUEST']._serialized_start=1494
  _globals['_APPREQUESTV2REQUEST']._serialized_end=1532
  _globals['_APPREQUESTRESPONSE']._serialized_start=1534
  _globals['_APPREQUESTRESPONSE']._serialized_end=1572
  _globals['_APPPUBLISHREQUEST']._serialized_start=1574
  _globals['_APPPUBLISHREQUEST']._serialized_end=1610
  _globals['_DOWNLOADFILEREQUEST']._serialized_start=1612
  _globals['_DOWNLOADFILEREQUEST']._serialized_end=1660
  _globals['_APPDOWNLOADREQUEST']._serialized_start=1662
  _globals['_APPDOWNLOADREQUEST']._serialized_end=1727
  _globals['_APPDOWNLOADRESPONSE']._serialized_start=1729
  _globals['_APPDOWNLOADRESPONSE']._serialized_end=1764
  _globals['_APPUPLOADREQUEST']._serialized_start=1766
  _globals['_APPUPLOADREQUEST']._serialized_end=1810
  _globals['_APPUPLOADRESPONSE']._serialized_start=1812
  _globals['_APPUPLOADRESPONSE']._serialized_end=1844
  _globals['_GATEWAYREQUESTREQUEST']._serialized_start=1847
  _globals['_GATEWAYREQUESTREQUEST']._serialized_end=2026
  _globals['_GATEWAYREQUESTREQUEST_HEADERSENTRY']._serialized_start=1980
  _globals['_GATEWAYREQUESTREQUEST_HEADERSENTRY']._serialized_end=2026
  _globals['_GATEWAYREQUESTRESPONSE']._serialized_start=2029
  _globals['_GATEWAYREQUESTRESPONSE']._serialized_end=2196
  _globals['_GATEWAYREQUESTRESPONSE_HEADERSENTRY']._serialized_start=1980
  _globals['_GATEWAYREQUESTRESPONSE_HEADERSENTRY']._serialized_end=2026
  _globals['_HTTPREQUEST']._serialized_start=2199
  _globals['_HTTPREQUEST']._serialized_end=2353
  _globals['_HTTPREQUEST_HEADERSENTRY']._serialized_start=1980
  _globals['_HTTPREQUEST_HEADERSENTRY']._serialized_end=2026
  _globals['_HTTPRESPONSE']._serialized_start=2356
  _globals['_HTTPRESPONSE']._serialized_end=2504
  _globals['_HTTPRESPONSE_HEADERSENTRY']._serialized_start=1980
  _globals['_HTTPRESPONSE_HEADERSENTRY']._serialized_end=2026
  _globals['_NODEINFO']._serialized_start=2506
  _globals['_NODEINFO']._serialized_end=2563
  _globals['_GETPORTCONNECTIONSREQUEST']._serialized_start=2565
  _globals['_GETPORTCONNECTIONSREQUEST']._serialized_end=2620
  _globals['_GETPORTCONNECTIONSRESPONSE']._serialized_start=2622
  _globals['_GETPORTCONNECTIONSRESPONSE']._serialized_end=2682
  _globals['_NODE']._serialized_start=2685
  _globals['_NODE']._serialized_end=2982
  _globals['_RUNTIMEHELPER']._serialized_start=2985
  _globals['_RUNTIMEHELPER']._serialized_end=4294
# @@protoc_insertion_point(module_scope)



import json
from jsonpath_ng import jsonpath, parse
from robomotion.capabilities import IsLMOCapable
from robomotion.lmo import IsLMO, SerializeLMO, UnpackMessageBytes, PackMessageBytes, DeserializeFromDict
class Context:
    def get_id(self) -> str:
        pass

    def set(self, key: str, value: object):
        pass

    def get(self, key: str) -> object:
        pass

    def get_raw(self) -> bytes:
        pass

    def set_raw(self, data: bytes):
        pass

    def is_empty(self) -> bool:
        pass

    def clone(self) -> 'Context':
        """Create and return a new Context instance with the same data."""
        return Context()

class Message(Context):
    def __init__(self, data: bytes):
        msg = json.loads(data.decode("utf-8"))
        self.id = msg["id"]
        self.data = data

    def get_id(self) -> str:
        return str(self.id)

    def clone(self) -> 'Message':
        """Create and return a new Message instance with the same data."""
        return Message(self.get_raw())

    def set(self, key: str, value: object):
        if IsLMOCapable():            
            serialized_value, err = SerializeLMO(value)
            if serialized_value:
                value = serialized_value                    
        msg = json.loads(self.data.decode("utf-8"))
        msg[key] = value
        self.data = json.dumps(msg).encode("utf-8")
    def get(self, key: str) -> object:
        msg = json.loads(self.data.decode("utf-8"))
        val = parse("$.%s" % key).find(msg)
        if len(val) == 0:
            return None
        result = val[0].value
        if  IsLMOCapable() and IsLMO(result):
            res = DeserializeFromDict(result)  
            if res:
                result = res.Value()
            else: 
                return None            
        return result
    
    def delete(self, key: str):
        msg = json.loads(self.data.decode("utf-8"))
        del msg[key]
        self.data = json.dumps(msg).encode("utf-8")

    def get_raw(self) -> bytes:
        return self.data

    def set_raw(self, data: bytes):
        self.data = data

    def is_empty(self) -> bool:
        return self.data == None or len(self.data) == 0



from robomotion.runtime import Runtime
from robomotion.error import RuntimeNotInitializedError
from robomotion.plugin_pb2 import (
    DebugRequest,
    EmitOutputRequest,
    EmitInputRequest,
    EmitErrorRequest,
    EmitFlowEventRequest,
    AppRequestRequest,
)
import pickle

class Event:
    @staticmethod
    def emit_debug(guid: str, name: str, message: object):
        if Runtime.client == None:
            raise RuntimeNotInitializedError

        request = DebugRequest(guid=guid, name=name, message=pickle.dumps(message))
        Runtime.client.Debug(request)

    @staticmethod
    def emit_output(guid: str, output: bytes, port: int):
        if Runtime.client == None:
            raise RuntimeNotInitializedError

        request = EmitOutputRequest(guid=guid, output=output, port=port)
        Runtime.client.EmitOutput(request)

    @staticmethod
    def emit_input(guid: str, input: bytes):
        if Runtime.client == None:
            raise RuntimeNotInitializedError

        request = EmitInputRequest(guid=guid, input=input)
        Runtime.client.EmitInput(request)

    @staticmethod
    def emit_error(guid: str, name: str, message: str):
        if Runtime.client == None:
            raise RuntimeNotInitializedError

        request = EmitErrorRequest(guid=guid, name=name, message=message)
        Runtime.client.EmitError(request)

    @staticmethod
    def emit_flow_event(guid: str, name: str):
        if Runtime.client == None:
            raise RuntimeNotInitializedError

        request = EmitFlowEventRequest(guid=guid, name=name)
        Runtime.client.EmitFlowEvent(request)

    @staticmethod
    def app_request(data: bytes, timeout: int):
        if Runtime.client == None:
            raise RuntimeNotInitializedError

        request = AppRequestRequest(request=data, timeout=timeout)
        Runtime.client.AppRequest(request)



from cmath import isnan
from distutils.log import debug
from enum import Enum
from select import select
import grpc
from json import dumps, loads
import os
from psutil import net_connections, process_iter
from psutil._common import CONN_LISTEN
from robomotion.utils import File
from robomotion import runner_pb2, runner_pb2_grpc


class AttachConfig:
    def __init__(self, protocol: str, addr: str, pid: int, namespace: str):
        self.protocol = protocol
        self.addr = addr
        self.pid = pid
        self.namespace = namespace


class EProtocol(Enum):
    ProtocolInvalid = ""
    ProtocolNetRPC = "netrpc"
    ProtocolGRPC = "grpc"


class Debug:
    atteched_to: str = ""

    @staticmethod
    def attach(g_addr: str, ns: str):
        cfg = AttachConfig(str(EProtocol.ProtocolGRPC.value), g_addr, os.getpid(), ns)
        cfg_data = dumps(cfg.__dict__)

        Debug.atteched_to = Debug.get_rpc_addr()
        if Debug.atteched_to == "":
            print("empty gRPC address")
            exit(0)

        print("Attached to %s" % Debug.atteched_to)

        channel = grpc.insecure_channel(Debug.atteched_to)
        client = runner_pb2_grpc.DebugStub(channel)

        request = runner_pb2.AttachRequest(config=cfg_data.encode())
        client.Attach(request)
        channel.close()

    @staticmethod
    def detach(ns: str):
        if Debug.atteched_to == "":
            print("empty gRPC address")
            exit(0)

        channel = grpc.insecure_channel(Debug.atteched_to)
        client = runner_pb2_grpc.DebugStub(channel)

        request = runner_pb2.DetachRequest(namespace=ns)
        client.Detach(request)
        channel.close()

    @staticmethod
    def get_rpc_addr() -> str:
        tabs = Debug.get_netstat_ports(CONN_LISTEN, "robomotion-runner")
        tabs = Debug.filter_tabs(tabs)

        if len(tabs) == 0:
            return ""
        elif len(tabs) == 1:
            return "%s:%d" % (tabs[0]["sconn"].laddr.ip, tabs[0]["sconn"].laddr.port)

        return Debug.select_tab(tabs)

    @staticmethod
    def filter_tabs(tabs: list) -> list:
        filtered = []
        for tab in tabs:
            addr = "%s:%d" % (tab["sconn"].laddr.ip, tab["sconn"].laddr.port)
            try:
                tab["robot_name"] = Debug.get_robot_name(addr)
                filtered.append(tab)
            except Exception as e:
                print(e)

        return filtered

    @staticmethod
    def get_netstat_ports(state: str = CONN_LISTEN, process_name: str = ""):
        tabs = net_connections()

        pids = []
        for proc in process_iter():
            if process_name.lower() in proc.name().lower():
                pids.append(proc.pid)

        filtered = filter(lambda tab: tab.status == state and tab.pid in pids, tabs)
        return list(map(lambda tab: {"sconn": tab}, filtered))

    @staticmethod
    def select_tab(tabs: list) -> str:
        count = len(tabs)

        robots = ""
        for i in range(0, count):
            robots += "%d) %s\n" % (i + 1, tabs[i]["robot_name"])

        selected = 0
        print("Found %d robots running on the machine:" % count)
        print("%s" % robots)
        print("Please select a robot to attach (1-%d):" % count)

        while True:
            line = input()
            if not line.isnumeric():
                continue

            selected = int(line)
            if selected > 0 and selected <= count:
                tab = tabs[selected - 1]
                return "%s:%d" % (tab["sconn"].laddr.ip, tab["sconn"].laddr.port)

    @staticmethod
    def get_robot_name(addr: str) -> str:
        channel = grpc.insecure_channel(addr)
        client = runner_pb2_grpc.RunnerStub(channel)

        request = runner_pb2.Null()
        resp = client.RobotName(request)
        channel.close()
        return resp.robot_name



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc
import warnings

import plugin_pb2 as plugin__pb2

GRPC_GENERATED_VERSION = '1.64.1'
GRPC_VERSION = grpc.__version__
EXPECTED_ERROR_RELEASE = '1.65.0'
SCHEDULED_RELEASE_DATE = 'June 25, 2024'
_version_not_supported = False

try:
    from grpc._utilities import first_version_is_lower
    _version_not_supported = first_version_is_lower(GRPC_VERSION, GRPC_GENERATED_VERSION)
except ImportError:
    _version_not_supported = True

if _version_not_supported:
    warnings.warn(
        f'The grpc package installed is at version {GRPC_VERSION},'
        + f' but the generated code in plugin_pb2_grpc.py depends on'
        + f' grpcio>={GRPC_GENERATED_VERSION}.'
        + f' Please upgrade your grpc module to grpcio>={GRPC_GENERATED_VERSION}'
        + f' or downgrade your generated code using grpcio-tools<={GRPC_VERSION}.'
        + f' This warning will become an error in {EXPECTED_ERROR_RELEASE},'
        + f' scheduled for release on {SCHEDULED_RELEASE_DATE}.',
        RuntimeWarning
    )


class NodeStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.Init = channel.unary_unary(
                '/proto.Node/Init',
                request_serializer=plugin__pb2.InitRequest.SerializeToString,
                response_deserializer=plugin__pb2.Empty.FromString,
                _registered_method=True)
        self.OnCreate = channel.unary_unary(
                '/proto.Node/OnCreate',
                request_serializer=plugin__pb2.OnCreateRequest.SerializeToString,
                response_deserializer=plugin__pb2.OnCreateResponse.FromString,
                _registered_method=True)
        self.OnMessage = channel.unary_unary(
                '/proto.Node/OnMessage',
                request_serializer=plugin__pb2.OnMessageRequest.SerializeToString,
                response_deserializer=plugin__pb2.OnMessageResponse.FromString,
                _registered_method=True)
        self.OnClose = channel.unary_unary(
                '/proto.Node/OnClose',
                request_serializer=plugin__pb2.OnCloseRequest.SerializeToString,
                response_deserializer=plugin__pb2.OnCloseResponse.FromString,
                _registered_method=True)
        self.GetCapabilities = channel.unary_unary(
                '/proto.Node/GetCapabilities',
                request_serializer=plugin__pb2.Empty.SerializeToString,
                response_deserializer=plugin__pb2.PGetCapabilitiesResponse.FromString,
                _registered_method=True)


class NodeServicer(object):
    """Missing associated documentation comment in .proto file."""

    def Init(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def OnCreate(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def OnMessage(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def OnClose(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetCapabilities(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_NodeServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'Init': grpc.unary_unary_rpc_method_handler(
                    servicer.Init,
                    request_deserializer=plugin__pb2.InitRequest.FromString,
                    response_serializer=plugin__pb2.Empty.SerializeToString,
            ),
            'OnCreate': grpc.unary_unary_rpc_method_handler(
                    servicer.OnCreate,
                    request_deserializer=plugin__pb2.OnCreateRequest.FromString,
                    response_serializer=plugin__pb2.OnCreateResponse.SerializeToString,
            ),
            'OnMessage': grpc.unary_unary_rpc_method_handler(
                    servicer.OnMessage,
                    request_deserializer=plugin__pb2.OnMessageRequest.FromString,
                    response_serializer=plugin__pb2.OnMessageResponse.SerializeToString,
            ),
            'OnClose': grpc.unary_unary_rpc_method_handler(
                    servicer.OnClose,
                    request_deserializer=plugin__pb2.OnCloseRequest.FromString,
                    response_serializer=plugin__pb2.OnCloseResponse.SerializeToString,
            ),
            'GetCapabilities': grpc.unary_unary_rpc_method_handler(
                    servicer.GetCapabilities,
                    request_deserializer=plugin__pb2.Empty.FromString,
                    response_serializer=plugin__pb2.PGetCapabilitiesResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'proto.Node', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers('proto.Node', rpc_method_handlers)


 # This class is part of an EXPERIMENTAL API.
class Node(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def Init(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/proto.Node/Init',
            plugin__pb2.InitRequest.SerializeToString,
            plugin__pb2.Empty.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def OnCreate(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/proto.Node/OnCreate',
            plugin__pb2.OnCreateRequest.SerializeToString,
            plugin__pb2.OnCreateResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def OnMessage(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/proto.Node/OnMessage',
            plugin__pb2.OnMessageRequest.SerializeToString,
            plugin__pb2.OnMessageResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def OnClose(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/proto.Node/OnClose',
            plugin__pb2.OnCloseRequest.SerializeToString,
            plugin__pb2.OnCloseResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetCapabilities(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/proto.Node/GetCapabilities',
            plugin__pb2.Empty.SerializeToString,
            plugin__pb2.PGetCapabilitiesResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)


class RuntimeHelperStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.Close = channel.unary_unary(
                '/proto.RuntimeHelper/Close',
                request_serializer=plugin__pb2.Empty.SerializeToString,
                response_deserializer=plugin__pb2.Empty.FromString,
                _registered_method=True)
        self.Debug = channel.unary_unary(
                '/proto.RuntimeHelper/Debug',
                request_serializer=plugin__pb2.DebugRequest.SerializeToString,
                response_deserializer=plugin__pb2.Empty.FromString,
                _registered_method=True)
        self.EmitFlowEvent = channel.unary_unary(
                '/proto.RuntimeHelper/EmitFlowEvent',
                request_serializer=plugin__pb2.EmitFlowEventRequest.SerializeToString,
                response_deserializer=plugin__pb2.Empty.FromString,
                _registered_method=True)
        self.EmitInput = channel.unary_unary(
                '/proto.RuntimeHelper/EmitInput',
                request_serializer=plugin__pb2.EmitInputRequest.SerializeToString,
                response_deserializer=plugin__pb2.Empty.FromString,
                _registered_method=True)
        self.EmitOutput = channel.unary_unary(
                '/proto.RuntimeHelper/EmitOutput',
                request_serializer=plugin__pb2.EmitOutputRequest.SerializeToString,
                response_deserializer=plugin__pb2.Empty.FromString,
                _registered_method=True)
        self.EmitError = channel.unary_unary(
                '/proto.RuntimeHelper/EmitError',
                request_serializer=plugin__pb2.EmitErrorRequest.SerializeToString,
                response_deserializer=plugin__pb2.Empty.FromString,
                _registered_method=True)
        self.GetVaultItem = channel.unary_unary(
                '/proto.RuntimeHelper/GetVaultItem',
                request_serializer=plugin__pb2.GetVaultItemRequest.SerializeToString,
                response_deserializer=plugin__pb2.GetVaultItemResponse.FromString,
                _registered_method=True)
        self.SetVaultItem = channel.unary_unary(
                '/proto.RuntimeHelper/SetVaultItem',
                request_serializer=plugin__pb2.SetVaultItemRequest.SerializeToString,
                response_deserializer=plugin__pb2.SetVaultItemResponse.FromString,
                _registered_method=True)
        self.GetVariable = channel.unary_unary(
                '/proto.RuntimeHelper/GetVariable',
                request_serializer=plugin__pb2.GetVariableRequest.SerializeToString,
                response_deserializer=plugin__pb2.GetVariableResponse.FromString,
                _registered_method=True)
        self.SetVariable = channel.unary_unary(
                '/proto.RuntimeHelper/SetVariable',
                request_serializer=plugin__pb2.SetVariableRequest.SerializeToString,
                response_deserializer=plugin__pb2.Empty.FromString,
                _registered_method=True)
        self.GetRobotInfo = channel.unary_unary(
                '/proto.RuntimeHelper/GetRobotInfo',
                request_serializer=plugin__pb2.Empty.SerializeToString,
                response_deserializer=plugin__pb2.GetRobotInfoResponse.FromString,
                _registered_method=True)
        self.AppRequest = channel.unary_unary(
                '/proto.RuntimeHelper/AppRequest',
                request_serializer=plugin__pb2.AppRequestRequest.SerializeToString,
                response_deserializer=plugin__pb2.AppRequestResponse.FromString,
                _registered_method=True)
        self.AppRequestV2 = channel.unary_unary(
                '/proto.RuntimeHelper/AppRequestV2',
                request_serializer=plugin__pb2.AppRequestV2Request.SerializeToString,
                response_deserializer=plugin__pb2.AppRequestResponse.FromString,
                _registered_method=True)
        self.AppPublish = channel.unary_unary(
                '/proto.RuntimeHelper/AppPublish',
                request_serializer=plugin__pb2.AppPublishRequest.SerializeToString,
                response_deserializer=plugin__pb2.Empty.FromString,
                _registered_method=True)
        self.DownloadFile = channel.unary_unary(
                '/proto.RuntimeHelper/DownloadFile',
                request_serializer=plugin__pb2.DownloadFileRequest.SerializeToString,
                response_deserializer=plugin__pb2.Empty.FromString,
                _registered_method=True)
        self.AppDownload = channel.unary_unary(
                '/proto.RuntimeHelper/AppDownload',
                request_serializer=plugin__pb2.AppDownloadRequest.SerializeToString,
                response_deserializer=plugin__pb2.AppDownloadResponse.FromString,
                _registered_method=True)
        self.AppUpload = channel.unary_unary(
                '/proto.RuntimeHelper/AppUpload',
                request_serializer=plugin__pb2.AppUploadRequest.SerializeToString,
                response_deserializer=plugin__pb2.AppUploadResponse.FromString,
                _registered_method=True)
        self.GatewayRequest = channel.unary_unary(
                '/proto.RuntimeHelper/GatewayRequest',
                request_serializer=plugin__pb2.GatewayRequestRequest.SerializeToString,
                response_deserializer=plugin__pb2.GatewayRequestResponse.FromString,
                _registered_method=True)
        self.ProxyRequest = channel.unary_unary(
                '/proto.RuntimeHelper/ProxyRequest',
                request_serializer=plugin__pb2.HttpRequest.SerializeToString,
                response_deserializer=plugin__pb2.HttpResponse.FromString,
                _registered_method=True)
        self.GetPortConnections = channel.unary_unary(
                '/proto.RuntimeHelper/GetPortConnections',
                request_serializer=plugin__pb2.GetPortConnectionsRequest.SerializeToString,
                response_deserializer=plugin__pb2.GetPortConnectionsResponse.FromString,
                _registered_method=True)
        self.IsRunning = channel.unary_unary(
                '/proto.RuntimeHelper/IsRunning',
                request_serializer=plugin__pb2.Empty.SerializeToString,
                response_deserializer=plugin__pb2.IsRunningResponse.FromString,
                _registered_method=True)


class RuntimeHelperServicer(object):
    """Missing associated documentation comment in .proto file."""

    def Close(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Debug(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def EmitFlowEvent(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def EmitInput(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def EmitOutput(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def EmitError(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetVaultItem(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SetVaultItem(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetVariable(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SetVariable(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetRobotInfo(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def AppRequest(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def AppRequestV2(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def AppPublish(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def DownloadFile(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def AppDownload(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def AppUpload(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GatewayRequest(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ProxyRequest(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetPortConnections(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def IsRunning(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_RuntimeHelperServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'Close': grpc.unary_unary_rpc_method_handler(
                    servicer.Close,
                    request_deserializer=plugin__pb2.Empty.FromString,
                    response_serializer=plugin__pb2.Empty.SerializeToString,
            ),
            'Debug': grpc.unary_unary_rpc_method_handler(
                    servicer.Debug,
                    request_deserializer=plugin__pb2.DebugRequest.FromString,
                    response_serializer=plugin__pb2.Empty.SerializeToString,
            ),
            'EmitFlowEvent': grpc.unary_unary_rpc_method_handler(
                    servicer.EmitFlowEvent,
                    request_deserializer=plugin__pb2.EmitFlowEventRequest.FromString,
                    response_serializer=plugin__pb2.Empty.SerializeToString,
            ),
            'EmitInput': grpc.unary_unary_rpc_method_handler(
                    servicer.EmitInput,
                    request_deserializer=plugin__pb2.EmitInputRequest.FromString,
                    response_serializer=plugin__pb2.Empty.SerializeToString,
            ),
            'EmitOutput': grpc.unary_unary_rpc_method_handler(
                    servicer.EmitOutput,
                    request_deserializer=plugin__pb2.EmitOutputRequest.FromString,
                    response_serializer=plugin__pb2.Empty.SerializeToString,
            ),
            'EmitError': grpc.unary_unary_rpc_method_handler(
                    servicer.EmitError,
                    request_deserializer=plugin__pb2.EmitErrorRequest.FromString,
                    response_serializer=plugin__pb2.Empty.SerializeToString,
            ),
            'GetVaultItem': grpc.unary_unary_rpc_method_handler(
                    servicer.GetVaultItem,
                    request_deserializer=plugin__pb2.GetVaultItemRequest.FromString,
                    response_serializer=plugin__pb2.GetVaultItemResponse.SerializeToString,
            ),
            'SetVaultItem': grpc.unary_unary_rpc_method_handler(
                    servicer.SetVaultItem,
                    request_deserializer=plugin__pb2.SetVaultItemRequest.FromString,
                    response_serializer=plugin__pb2.SetVaultItemResponse.SerializeToString,
            ),
            'GetVariable': grpc.unary_unary_rpc_method_handler(
                    servicer.GetVariable,
                    request_deserializer=plugin__pb2.GetVariableRequest.FromString,
                    response_serializer=plugin__pb2.GetVariableResponse.SerializeToString,
            ),
            'SetVariable': grpc.unary_unary_rpc_method_handler(
                    servicer.SetVariable,
                    request_deserializer=plugin__pb2.SetVariableRequest.FromString,
                    response_serializer=plugin__pb2.Empty.SerializeToString,
            ),
            'GetRobotInfo': grpc.unary_unary_rpc_method_handler(
                    servicer.GetRobotInfo,
                    request_deserializer=plugin__pb2.Empty.FromString,
                    response_serializer=plugin__pb2.GetRobotInfoResponse.SerializeToString,
            ),
            'AppRequest': grpc.unary_unary_rpc_method_handler(
                    servicer.AppRequest,
                    request_deserializer=plugin__pb2.AppRequestRequest.FromString,
                    response_serializer=plugin__pb2.AppRequestResponse.SerializeToString,
            ),
            'AppRequestV2': grpc.unary_unary_rpc_method_handler(
                    servicer.AppRequestV2,
                    request_deserializer=plugin__pb2.AppRequestV2Request.FromString,
                    response_serializer=plugin__pb2.AppRequestResponse.SerializeToString,
            ),
            'AppPublish': grpc.unary_unary_rpc_method_handler(
                    servicer.AppPublish,
                    request_deserializer=plugin__pb2.AppPublishRequest.FromString,
                    response_serializer=plugin__pb2.Empty.SerializeToString,
            ),
            'DownloadFile': grpc.unary_unary_rpc_method_handler(
                    servicer.DownloadFile,
                    request_deserializer=plugin__pb2.DownloadFileRequest.FromString,
                    response_serializer=plugin__pb2.Empty.SerializeToString,
            ),
            'AppDownload': grpc.unary_unary_rpc_method_handler(
                    servicer.AppDownload,
                    request_deserializer=plugin__pb2.AppDownloadRequest.FromString,
                    response_serializer=plugin__pb2.AppDownloadResponse.SerializeToString,
            ),
            'AppUpload': grpc.unary_unary_rpc_method_handler(
                    servicer.AppUpload,
                    request_deserializer=plugin__pb2.AppUploadRequest.FromString,
                    response_serializer=plugin__pb2.AppUploadResponse.SerializeToString,
            ),
            'GatewayRequest': grpc.unary_unary_rpc_method_handler(
                    servicer.GatewayRequest,
                    request_deserializer=plugin__pb2.GatewayRequestRequest.FromString,
                    response_serializer=plugin__pb2.GatewayRequestResponse.SerializeToString,
            ),
            'ProxyRequest': grpc.unary_unary_rpc_method_handler(
                    servicer.ProxyRequest,
                    request_deserializer=plugin__pb2.HttpRequest.FromString,
                    response_serializer=plugin__pb2.HttpResponse.SerializeToString,
            ),
            'GetPortConnections': grpc.unary_unary_rpc_method_handler(
                    servicer.GetPortConnections,
                    request_deserializer=plugin__pb2.GetPortConnectionsRequest.FromString,
                    response_serializer=plugin__pb2.GetPortConnectionsResponse.SerializeToString,
            ),
            'IsRunning': grpc.unary_unary_rpc_method_handler(
                    servicer.IsRunning,
                    request_deserializer=plugin__pb2.Empty.FromString,
                    response_serializer=plugin__pb2.IsRunningResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'proto.RuntimeHelper', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers('proto.RuntimeHelper', rpc_method_handlers)


 # This class is part of an EXPERIMENTAL API.
class RuntimeHelper(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def Close(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/proto.RuntimeHelper/Close',
            plugin__pb2.Empty.SerializeToString,
            plugin__pb2.Empty.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def Debug(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/proto.RuntimeHelper/Debug',
            plugin__pb2.DebugRequest.SerializeToString,
            plugin__pb2.Empty.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def EmitFlowEvent(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/proto.RuntimeHelper/EmitFlowEvent',
            plugin__pb2.EmitFlowEventRequest.SerializeToString,
            plugin__pb2.Empty.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def EmitInput(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/proto.RuntimeHelper/EmitInput',
            plugin__pb2.EmitInputRequest.SerializeToString,
            plugin__pb2.Empty.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def EmitOutput(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/proto.RuntimeHelper/EmitOutput',
            plugin__pb2.EmitOutputRequest.SerializeToString,
            plugin__pb2.Empty.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def EmitError(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/proto.RuntimeHelper/EmitError',
            plugin__pb2.EmitErrorRequest.SerializeToString,
            plugin__pb2.Empty.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetVaultItem(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/proto.RuntimeHelper/GetVaultItem',
            plugin__pb2.GetVaultItemRequest.SerializeToString,
            plugin__pb2.GetVaultItemResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def SetVaultItem(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/proto.RuntimeHelper/SetVaultItem',
            plugin__pb2.SetVaultItemRequest.SerializeToString,
            plugin__pb2.SetVaultItemResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetVariable(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/proto.RuntimeHelper/GetVariable',
            plugin__pb2.GetVariableRequest.SerializeToString,
            plugin__pb2.GetVariableResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def SetVariable(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/proto.RuntimeHelper/SetVariable',
            plugin__pb2.SetVariableRequest.SerializeToString,
            plugin__pb2.Empty.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetRobotInfo(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/proto.RuntimeHelper/GetRobotInfo',
            plugin__pb2.Empty.SerializeToString,
            plugin__pb2.GetRobotInfoResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def AppRequest(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/proto.RuntimeHelper/AppRequest',
            plugin__pb2.AppRequestRequest.SerializeToString,
            plugin__pb2.AppRequestResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def AppRequestV2(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/proto.RuntimeHelper/AppRequestV2',
            plugin__pb2.AppRequestV2Request.SerializeToString,
            plugin__pb2.AppRequestResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def AppPublish(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/proto.RuntimeHelper/AppPublish',
            plugin__pb2.AppPublishRequest.SerializeToString,
            plugin__pb2.Empty.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def DownloadFile(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/proto.RuntimeHelper/DownloadFile',
            plugin__pb2.DownloadFileRequest.SerializeToString,
            plugin__pb2.Empty.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def AppDownload(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/proto.RuntimeHelper/AppDownload',
            plugin__pb2.AppDownloadRequest.SerializeToString,
            plugin__pb2.AppDownloadResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def AppUpload(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/proto.RuntimeHelper/AppUpload',
            plugin__pb2.AppUploadRequest.SerializeToString,
            plugin__pb2.AppUploadResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GatewayRequest(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/proto.RuntimeHelper/GatewayRequest',
            plugin__pb2.GatewayRequestRequest.SerializeToString,
            plugin__pb2.GatewayRequestResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def ProxyRequest(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/proto.RuntimeHelper/ProxyRequest',
            plugin__pb2.HttpRequest.SerializeToString,
            plugin__pb2.HttpResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetPortConnections(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/proto.RuntimeHelper/GetPortConnections',
            plugin__pb2.GetPortConnectionsRequest.SerializeToString,
            plugin__pb2.GetPortConnectionsResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def IsRunning(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/proto.RuntimeHelper/IsRunning',
            plugin__pb2.Empty.SerializeToString,
            plugin__pb2.IsRunningResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from robomotion import runner_pb2 as runner__pb2


class RunnerStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.Init = channel.unary_unary(
            "/proto.Runner/Init",
            request_serializer=runner__pb2.InitRunnerRequest.SerializeToString,
            response_deserializer=runner__pb2.Null.FromString,
        )
        self.Run = channel.unary_unary(
            "/proto.Runner/Run",
            request_serializer=runner__pb2.RunRequest.SerializeToString,
            response_deserializer=runner__pb2.Null.FromString,
        )
        self.Clear = channel.unary_unary(
            "/proto.Runner/Clear",
            request_serializer=runner__pb2.Null.SerializeToString,
            response_deserializer=runner__pb2.Null.FromString,
        )
        self.RobotName = channel.unary_unary(
            "/proto.Runner/RobotName",
            request_serializer=runner__pb2.Null.SerializeToString,
            response_deserializer=runner__pb2.RobotNameResponse.FromString,
        )


class RunnerServicer(object):
    """Missing associated documentation comment in .proto file."""

    def Init(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def Run(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def Clear(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def RobotName(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")


def add_RunnerServicer_to_server(servicer, server):
    rpc_method_handlers = {
        "Init": grpc.unary_unary_rpc_method_handler(
            servicer.Init,
            request_deserializer=runner__pb2.InitRunnerRequest.FromString,
            response_serializer=runner__pb2.Null.SerializeToString,
        ),
        "Run": grpc.unary_unary_rpc_method_handler(
            servicer.Run,
            request_deserializer=runner__pb2.RunRequest.FromString,
            response_serializer=runner__pb2.Null.SerializeToString,
        ),
        "Clear": grpc.unary_unary_rpc_method_handler(
            servicer.Clear,
            request_deserializer=runner__pb2.Null.FromString,
            response_serializer=runner__pb2.Null.SerializeToString,
        ),
        "RobotName": grpc.unary_unary_rpc_method_handler(
            servicer.RobotName,
            request_deserializer=runner__pb2.Null.FromString,
            response_serializer=runner__pb2.RobotNameResponse.SerializeToString,
        ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
        "proto.Runner", rpc_method_handlers
    )
    server.add_generic_rpc_handlers((generic_handler,))


# This class is part of an EXPERIMENTAL API.
class Runner(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def Init(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/proto.Runner/Init",
            runner__pb2.InitRunnerRequest.SerializeToString,
            runner__pb2.Null.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def Run(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/proto.Runner/Run",
            runner__pb2.RunRequest.SerializeToString,
            runner__pb2.Null.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def Clear(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/proto.Runner/Clear",
            runner__pb2.Null.SerializeToString,
            runner__pb2.Null.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def RobotName(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/proto.Runner/RobotName",
            runner__pb2.Null.SerializeToString,
            runner__pb2.RobotNameResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )


class RobotHelperStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """


class RobotHelperServicer(object):
    """Missing associated documentation comment in .proto file."""


def add_RobotHelperServicer_to_server(servicer, server):
    rpc_method_handlers = {}
    generic_handler = grpc.method_handlers_generic_handler(
        "proto.RobotHelper", rpc_method_handlers
    )
    server.add_generic_rpc_handlers((generic_handler,))


# This class is part of an EXPERIMENTAL API.
class RobotHelper(object):
    """Missing associated documentation comment in .proto file."""


class DebugStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.Attach = channel.unary_unary(
            "/proto.Debug/Attach",
            request_serializer=runner__pb2.AttachRequest.SerializeToString,
            response_deserializer=runner__pb2.Nil.FromString,
        )
        self.Detach = channel.unary_unary(
            "/proto.Debug/Detach",
            request_serializer=runner__pb2.DetachRequest.SerializeToString,
            response_deserializer=runner__pb2.Nil.FromString,
        )


class DebugServicer(object):
    """Missing associated documentation comment in .proto file."""

    def Attach(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def Detach(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")


def add_DebugServicer_to_server(servicer, server):
    rpc_method_handlers = {
        "Attach": grpc.unary_unary_rpc_method_handler(
            servicer.Attach,
            request_deserializer=runner__pb2.AttachRequest.FromString,
            response_serializer=runner__pb2.Nil.SerializeToString,
        ),
        "Detach": grpc.unary_unary_rpc_method_handler(
            servicer.Detach,
            request_deserializer=runner__pb2.DetachRequest.FromString,
            response_serializer=runner__pb2.Nil.SerializeToString,
        ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
        "proto.Debug", rpc_method_handlers
    )
    server.add_generic_rpc_handlers((generic_handler,))


# This class is part of an EXPERIMENTAL API.
class Debug(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def Attach(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/proto.Debug/Attach",
            runner__pb2.AttachRequest.SerializeToString,
            runner__pb2.Nil.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def Detach(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/proto.Debug/Detach",
            runner__pb2.DetachRequest.SerializeToString,
            runner__pb2.Nil.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )



import logging
import sys
import robomotion.runtime  as runtime
from robomotion.utils import Version

MINIMUM_ROBOT_VERSION = "23.12.0"
class Capability:
    CapabilityLMO = 1 << 0

capabilities = []

def IsLMOCapable():
    robot_info = runtime.Runtime.get_robot_info()  
    
    capabilities = robot_info.get("capabilities", {})    
    lmo_enabled = capabilities.get("lmo", False)
    
    version = robot_info.get("version", "")
    if lmo_enabled and not Version.is_version_less_than(version, MINIMUM_ROBOT_VERSION):
        return True

    return False

def add_capability(capability):
    capabilities.append(capability)


def get_capabilities():
    _capabilities = sys.maxsize 
    for cap in capabilities:
        _capabilities &= cap
    return _capabilities

def init_capabilities():
    add_capability(Capability.CapabilityLMO)
    
init_capabilities()


import json
import os
import base64
import uuid
import robomotion.runtime  as runtime
from robomotion.utils import get_temp_path
LMO_MAGIC = 0x1343B7E
LMO_LIMIT = 256 << 10  # 256kb
LMO_VERSION = 0x01
LMO_HEAD = 100

enableLMO = False

def SetLMOFag(flag):
    global enableLMO
    enableLMO = flag
def GetLMOFlag():
    return enableLMO

class LargeMessageObject:
    def __init__(self, id, head, size,data):
        self.Magic = LMO_MAGIC
        self.Version = LMO_VERSION
        self.ID = id
        self.Head = head 
        self.Size = size
        self.Data = data

    def Value(self):
        return self.Data
    
    def ToJSON(self):        
        return {        
            "magic":self.Magic,
            "version":self.Version,
            "id": self.ID,
            "head" : self.Head,
            "size": self.Size,
            "data": self.Data        
        }
    

def NewId() -> str:
    uid = uuid.uuid4().bytes
    encoded = base64.b32encode(uid).decode('utf-8')
    return encoded.replace('=', '')[:26]

def IsLMOCapable():
    return True # düzelt
def SerializeLMO(value):
    global LMO_MAGIC, LMO_VERSION, LMO_HEAD, LMO_LIMIT
    if not IsLMOCapable():
        return None
    data = json.dumps(value).encode()
    dataLen = len(data)

    if dataLen < LMO_LIMIT:
        return None
    
    id = NewId()
    head = data[:LMO_HEAD].decode()
    lmo = LargeMessageObject(id, head, dataLen, value)
    robot_info = runtime.Runtime.get_robot_info()    
    robotID = robot_info["id"]

    tempPath = get_temp_path()
    
    if not robotID or not tempPath:
        return None
    dir_path = os.path.join(tempPath, "robots", robotID)
    os.makedirs(dir_path, exist_ok=True, mode=0o755)

    file_path = os.path.join(dir_path, id + ".lmo")
    with open(file_path, "w") as file:
        lmo_json = json.dumps(lmo.__dict__)
        file.write(lmo_json)
    
    lmo.Data = None
    lmo = lmo.ToJSON()
    return lmo

def DeserializeLMO(id):
    robotInfo = runtime.Runtime.get_robot_info()    

    robotID = robotInfo["id"]
    tempPath = get_temp_path()

    if not robotID or not tempPath:
        return None

    dir_path = os.path.join(tempPath, "robots", robotID)
    file_path = os.path.join(dir_path, id + ".lmo")

    try:
        with open(file_path, "r") as file:
            file_content = file.read()
            lmo_dict =  json.loads(file_content)
            
            lmo = LargeMessageObject(lmo_dict["id"], lmo_dict["head"], lmo_dict["size"], lmo_dict["data"]) #check et
            return lmo
    except FileNotFoundError as e:
        return None
    
def DeserializeFromDict(d):
    if IsLMOCapable():
        id = d["id"]
        if not id:
            return None
        return DeserializeLMO(id)
    return None
def PackMessageBytes(inMsg):
    global LMO_LIMIT

    if not IsLMOCapable() or len(inMsg) < LMO_LIMIT:
        return inMsg, None

    try:
        msg = json.loads(inMsg.decode())
        msg = PackMessage(msg)
        packed = json.dumps(msg).encode()
        return packed, None
    except json.JSONDecodeError as e:
        return None, e

def PackMessage(msg):
    if not IsLMOCapable():
        return

    for key, value in msg.items():
        lmo, e = SerializeLMO(value)
        if e:
            return None

        if lmo:
            msg[key] = lmo
        else:
            msg[key] = value
        
    return msg

def UnpackMessageBytes(inMsg):
    try:
        msg = json.loads(inMsg.decode())
        UnpackMessage(inMsg, msg)
        return json.dumps(msg).encode(), None
    except json.JSONDecodeError as e:
        return None, e
    


def UnpackMessage(inMsg, msg):
    if not IsLMOCapable():
        return

    try:
        msg = json.loads(inMsg.decode())
        for key, value in msg.items():
            lmo = value

            if "magic" in lmo and lmo["magic"] == LMO_MAGIC:
                idValue = lmo.get("id")
                if idValue:
                    result, err = DeserializeLMO(idValue)
                    if err:
                        return err
                    msg[key] = result.Value()
    except AttributeError as e:
        return e

def IsLMO(value):
    global LMO_MAGIC

    if not IsLMOCapable() or type(value) != dict:
        return False

    return value.get("magic") == LMO_MAGIC

def DeleteLMObyID(id):
    robot_info = runtime.Runtime.get_robot_info()    

    robotID = robot_info["id"]
    tempPath = get_temp_path()

    if not robotID or not tempPath:
        return
    
    dir_path = os.path.join(tempPath, "robots", robotID, id)
    os.remove(dir_path)





from pathlib import Path
from sys import platform
from packaging import version
import os

class File:
    @staticmethod
    def temp_dir():
        home = str(Path.home())
        is_win = platform.startswith("win")
        if is_win:
            return "%s\\AppData\\Local\\Temp\\Robomotion" % home
        return "/tmp/robomotion"


class Version:
    @staticmethod
    def is_version_less_than(ver: str, other: str) -> bool:
        v = version.parse(ver)
        v2 = version.parse(other)
        return v < v2

def get_temp_path():
    if platform == "win32":
        home = os.getenv("HOMEDRIVE") + os.getenv("HOMEPATH")
        if home == "":
            home = os.getenv("USERPROFILE")
        return os.path.join(home, "AppData", "Local", "Robomotion", "temp")
    elif platform.startswith("linux") or platform == "darwin":
        return os.path.join(user_home_dir(), ".config", "robomotion", "temp")

    return ""

def user_home_dir():
    if platform == "win32":
        home = os.getenv("HOMEDRIVE") + os.getenv("HOMEPATH")
        if home == "":
            home = os.getenv("USERPROFILE")
        return home
    return os.getenv("HOME")


import inspect
import json
import sys
from robomotion.node import Node
from robomotion.variable import (
    Variable,
    InVariable,
    OptVariable,
    OutVariable,
    Credentials,
    ECategory,
)
from re import sub
from robomotion.port import Port
import base64


class NodeSpec:
    def __init__(
        self,
        id: str,
        icon: str = "",
        name: str = "",
        color: str = "",
        editor: str = None,
        inputs: int = 0,
        outputs: int = 0,
        filters: str = None,
        custom_ports: list = None,
    ):
        self.id = id
        self.icon = icon
        self.name = name
        self.color = color
        self.editor = editor
        self.inputs = inputs
        self.inputs = self.inputs[0]
        self.outputs = outputs
        self.outputs = self.outputs[0]
        self.filters = filters
        self.customPorts = custom_ports or []
        self.properties = []


class Property:
    def __init__(self, schema):
        self.schema = schema
        self.formData = {}
        self.uiSchema = {}


class Schema:
    def __init__(self, type: str, title: str):
        self.type = type
        self.title = title
        self.properties = {}


class SProperty:
    type = ""
    title = ""
    description = None
    subtitle = None
    category = None
    properties = None
    csScope = None
    jsScope = None
    customScope = None
    messageScope = None
    messageOnly = None
    multiple = None
    variableType = None
    enum = []
    enumNames = []


class VarDataProperty:
    def __init__(self, name: str, scope: str):
        self.scope = scope
        self.name = name


class Spec:
    @staticmethod
    def generate(plugin_name, version):
        frm = inspect.stack()[2]
        mod = inspect.getmodule(frm[0])
        clsmembers = inspect.getmembers(sys.modules[mod.__name__], inspect.isclass)

        pspec = {"name": plugin_name, "version": version}

        nodes = []

        for c in clsmembers:
            cls = c[1]
            if issubclass(cls, Node) and cls is not Node:
                node = {}
                inst = cls()
                node["id"] = inst.name
                node["name"] = inst.title
                node["icon"] = inst.icon
                node["color"] = inst.color
                node["editor"] = inst.editor
                node["inputs"] = inst.inputs
                node["outputs"] = inst.outputs

                # --- customPorts support ---
                custom_ports = []
                for attr in dir(cls):
                    value = getattr(cls, attr, None)
                    if isinstance(value, Port):
                        port_spec = {
                            "direction": value.direction,
                            "position": value.position,
                            "name": value.name,
                            "icon": value.icon,
                            "color": value.color,
                        }
                        if value.filters:
                            port_spec["filters"] = value.filters
                        # Attach order for sorting, but do not include in output
                        port_spec["_order"] = value.order if hasattr(value, "order") and value.order is not None else float('inf')
                        custom_ports.append(port_spec)
                # Sort by _order, then remove _order from output
                custom_ports.sort(key=lambda p: p["_order"])
                for p in custom_ports:
                    p.pop("_order", None)
                if custom_ports:
                    node["customPorts"] = custom_ports

                # --- filters support ---
                if hasattr(inst, "filters") and inst.filters:
                    node["filters"] = inst.filters

                properties = []
                inputVars = Spec.get_inputs_vars(cls)
                inputs = Spec.get_inputs(cls)

                # Merge and sort inputVars and inputs by order
                all_inputs = inputVars + inputs
                all_inputs = sorted(all_inputs, key=lambda x: getattr(x['val'], 'order', float('inf')) if hasattr(x['val'], 'order') and x['val'].order is not None else float('inf'))

                if len(all_inputs) > 0:
                    prop = {}

                    pSchema, pUISchema = {}, {}
                    pSchema["title"] = "Input"
                    pSchema["type"] = "object"

                    inProperties, formData = {}, {}
                    uiOrder = []

                    for _input in all_inputs:
                        input = _input["val"]
                        name = Spec.lower_first_letter(_input["key"])
                        # Special case for func: always output as type string

                        # --- b64: decoding for variable names ---
                        n = getattr(input, "name", "")
                        if isinstance(n, str) and n.startswith("b64:"):
                            try:
                                n = base64.b64decode(n[4:]).decode("utf-8")
                            except Exception:
                                pass

                        if name == "func":
                            inObject = {
                                "type": "string",
                                "title": input.title
                            }
                            pUISchema[name] = {"ui:widget": "hidden"}
                            inProperties[name] = inObject
                            formData[name] = n if name == "func" else {"scope": input.scope, "name": n} if isinstance(input, InVariable) else input.default
                        elif isinstance(input, InVariable):
                            inObject = {
                                "title": input.title,
                                "variableType": input.type,
                            }
                            if input.arrayFields is not None:
                                inObject["type"] = "array"
                                arrProps = {}
                                arr_fields = input.arrayFields
                                if isinstance(arr_fields, str):
                                    arr_fields = [arr_fields]
                                for arrField in arr_fields:
                                    arrProps[Spec.to_snake_case(arrField)] = {
                                        "type": "string",
                                        "title": arrField,
                                    }
                                inObject["items"] = {
                                    "type": "object",
                                    "properties": {
                                        "scope": {"type": "string"},
                                        "name": {"properties": arrProps},
                                    },
                                }
                            else:
                                inObject["type"] = "object"
                                inObject["properties"] = {
                                    "scope": {"type": "string"},
                                    "name": {"type": "string"},
                                }
                            if input.customScope:
                                inObject["customScope"] = True
                            if input.messageScope:
                                inObject["messageScope"] = True
                            if input.messageOnly:
                                inObject["messageOnly"] = True
                            if input.jsScope:
                                inObject["jsScope"] = True
                            if input.pyScope:
                                inObject["pyScope"] = True
                            if input.description != "":
                                inObject["description"] = input.description
                            if input.format != "":
                                inObject["format"] = input.format
                                pUISchema[name] = {"ui:field": input.format}
                            else:
                                if input.arrayFields is not None:
                                    pUISchema[name] = {"ui:field": "array"}
                                else:
                                    pUISchema[name] = {"ui:field": "variable"}
                            inProperties[name] = inObject
                            if input.arrayFields is not None:
                                formData[name] = []
                            else:
                                formData[name] = {"scope": input.scope, "name": input.name}
                        else:
                            inObject = {"type": input.type, "title": input.title}
                            if input.hidden:
                                pUISchema[name] = {"ui:widget": "hidden"}
                            if input.description != "":
                                inObject["description"] = input.description
                                pUISchema[name] = {"ui:field": "input"}
                            if input.format != "":
                                inObject["format"] = input.format
                                pUISchema[name] = {"ui:field": input.format}
                            inProperties[name] = inObject
                            formData[name] = input.default
                        uiOrder.append(name)

                    pSchema["properties"] = inProperties
                    pUISchema["ui:order"] = uiOrder

                    prop["schema"] = pSchema
                    prop["uiSchema"] = pUISchema
                    prop["formData"] = formData

                    properties.append(prop)

                outputVars = Spec.get_output_vars(cls)
                outputs = Spec.get_outputs(cls)

                # Merge and sort outputVars and outputs by order
                all_outputs = outputVars + outputs
                all_outputs = sorted(all_outputs, key=lambda x: getattr(x['val'], 'order', float('inf')) if hasattr(x['val'], 'order') and x['val'].order is not None else float('inf'))

                if len(all_outputs) > 0:
                    prop = {}

                    pSchema, pUISchema = {}, {}
                    pSchema["title"] = "Output"
                    pSchema["type"] = "object"

                    outProperties, formData = {}, {}
                    uiOrder = []

                    for _output in all_outputs:
                        output = _output["val"]
                        name = Spec.lower_first_letter(_output["key"])
                        if isinstance(output, OutVariable):
                            outObject = {
                                "type": "object",
                                "title": output.title,
                                "variableType": output.type,
                                "properties": {
                                    "scope": {"type": "string"},
                                    "name": {"type": "string"},
                                },
                            }
                            if output.messageScope:
                                outObject["messageScope"] = True
                            if output.messageOnly:
                                outObject["messageOnly"] = True
                            if output.description != "":
                                outObject["description"] = output.description
                            pUISchema[name] = {"ui:field": "variable"}
                            formData[name] = {"scope": output.scope, "name": output.name}
                        else:
                            outObject = {"type": output.type, "title": output.title}
                            if output.hidden:
                                pUISchema[name] = {"ui:widget": "hidden"}
                            if output.description != "":
                                outObject["description"] = output.description
                                pUISchema[name] = {"ui:field": "input"}
                            formData[name] = output.default
                        outProperties[name] = outObject
                        uiOrder.append(name)

                    pSchema["properties"] = outProperties
                    pUISchema["ui:order"] = uiOrder

                    prop["schema"] = pSchema
                    prop["uiSchema"] = pUISchema
                    prop["formData"] = formData

                    properties.append(prop)

                optionVars = Spec.get_option_vars(cls)
                options = Spec.get_options(cls)
                credentials = Spec.get_credentials(cls)

                # Merge and sort optionVars, options, and credentials by order
                all_options = optionVars + options + credentials
                all_options = sorted(all_options, key=lambda x: getattr(x['val'], 'order', float('inf')) if hasattr(x['val'], 'order') and x['val'].order is not None else float('inf'))

                if len(all_options) > 0:
                    prop = {}

                    pSchema, pUISchema = {}, {}
                    pSchema["title"] = "Options"
                    pSchema["type"] = "object"

                    optProperties, formData = {}, {}
                    uiOrder = []

                    for _option in all_options:
                        option = _option["val"]
                        name = Spec.lower_first_letter(_option["key"])
                        # OptVariable
                        if isinstance(option, OptVariable):
                            optObject = {
                                "title": option.title,
                                "variableType": option.type,
                            }
                            if option.arrayFields is not None:
                                optObject["type"] = "array"
                                arrProps = {}
                                arr_fields = option.arrayFields
                                if isinstance(arr_fields, str):
                                    arr_fields = [arr_fields]
                                for arrField in arr_fields:
                                    arrProps[Spec.to_snake_case(arrField)] = {
                                        "type": "string",
                                        "title": arrField,
                                    }
                                optObject["items"] = {
                                    "type": "object",
                                    "properties": {
                                        "scope": {"type": "string"},
                                        "name": {"properties": arrProps},
                                    },
                                }
                            else:
                                optObject["type"] = "object"
                                optObject["properties"] = {
                                    "scope": {"type": "string"},
                                    "name": {"type": "string"},
                                }
                            if option.customScope:
                                optObject["customScope"] = True
                            if option.messageScope:
                                optObject["messageScope"] = True
                            if option.messageOnly:
                                optObject["messageOnly"] = True
                            if option.jsScope:
                                optObject["jsScope"] = True
                            if option.pyScope:
                                optObject["pyScope"] = True
                            if option.description != "":
                                optObject["description"] = option.description
                            if option.format != "":
                                optObject["format"] = option.format
                                pUISchema[name] = {"ui:field": option.format}
                            else:
                                if option.arrayFields is not None:
                                    pUISchema[name] = {"ui:field": "array"}
                                else:
                                    pUISchema[name] = {"ui:field": "variable"}
                            optProperties[name] = optObject
                            if option.arrayFields is not None:
                                formData[name] = []
                            else:
                                formData[name] = {"scope": option.scope, "name": option.name}
                        # Credentials
                        elif isinstance(option, Credentials):
                            optObject = {
                                "type": "object",
                                "title": option.title,
                                "subtitle": option.title,
                                "category": int(option.category),
                                "customScope": True,
                                "messageScope": True,
                                "properties": {
                                    "scope": {
                                        "type": "string",
                                    },
                                    "name": {
                                        "type": "object",
                                        "properties": {
                                            "vaultId": {"type": "string"},
                                            "itemId": {"type": "string"},
                                        },
                                    },
                                },
                            }
                            if option.description != "":
                                optObject["description"] = option.description
                            pUISchema[name] = {"ui:field": "vault"}
                            formData[name] = {
                                "scope": "Custom",
                                "name": {"vaultId": "_", "itemId": "_"},
                            }
                            optProperties[name] = optObject
                        # Variable (plain option)
                        else:
                            optObject = {"type": option.type, "title": option.title}
                            category = option.category
                            if category != ECategory.Null:
                                optObject["category"] = int(category)
                            if option.description != "":
                                optObject["description"] = option.description
                                pUISchema[name] = {"ui:field": "input"}
                            if option.format != "":
                                optObject["format"] = option.format
                                pUISchema[name] = {"ui:field": option.format}
                            if option.enum is not None:
                                enums = option.enum.enums
                                if enums is not None and len(enums) > 0:
                                    enumNames = option.enum.enumNames
                                    optObject["enum"] = enums
                                    optObject["enumNames"] = enumNames
                            if option.hidden:
                                pUISchema[name] = {"ui:widget": "hidden"}
                            if option.default is not None:
                                formData[name] = option.default
                            optProperties[name] = optObject
                        uiOrder.append(name)

                    pSchema["properties"] = optProperties
                    pUISchema["ui:order"] = uiOrder

                    prop["schema"] = pSchema
                    prop["uiSchema"] = pUISchema
                    prop["formData"] = formData

                    properties.append(prop)

                node["properties"] = properties
                nodes.append(node)

        pspec["nodes"] = nodes
        print(json.dumps(pspec, indent=2))

    @staticmethod
    def cleandict(d):
        if not isinstance(d, dict):
            return d
        return dict((k, Spec.cleandict(v)) for k, v in d.iteritems() if v is not None)

    @staticmethod
    def jsonKeys2int(x):
        if isinstance(x, dict):
            return {int(k): v for k, v in x.items()}
        return x

    @staticmethod
    def get_variable_type(val) -> str:
        if isinstance(val, Variable):
            return Spec.upper_first_letter(val.type)

        return "String"

    @staticmethod
    def lower_first_letter(text: str) -> str:
        if len(text) < 2:
            return text.lower()
        return text[:1].lower() + text[1:]

    @staticmethod
    def upper_first_letter(text: str) -> str:
        if len(text) < 2:
            return text.upper()
        return text[:1].upper() + text[1:]

    @staticmethod
    def get_inputs(cls) -> list[any]:
        inputs = []
        fields = cls().cls().__dict__
        for key in fields.keys():
            val = fields[key]
            # Special case for inFunc
            if key == "inFunc":
                prop_key = "Func"
            else:
                prop_key = key
            if (
                isinstance(val, Variable)
                and not isinstance(val, InVariable)
                and not isinstance(val, OutVariable)
                and not isinstance(val, OptVariable)
                and val.input
            ):
                inputs.append({"key": prop_key, "val": val})
            elif (
                not isinstance(val, InVariable)
                and not isinstance(val, OutVariable)
                and not isinstance(val, OptVariable)
                and key.lower().startswith("in")
            ):
                inputs.append(
                    {
                        "key": prop_key,
                        "val": Variable(
                            title=Spec.camel_case_to_text(key.lstrip("in")),
                            type=Spec.get_type(val),
                            default=val,
                        ),
                    }
                )

        return inputs

    @staticmethod
    def get_inputs_vars(cls) -> list[any]:
        inputs = []
        fields = cls().cls().__dict__
        for key in fields.keys():
            val = fields[key]
            # Special case for inFunc
            if key == "inFunc":
                prop_key = "Func"
            else:
                prop_key = key
            if isinstance(val, InVariable):
                inputs.append({"key": prop_key, "val": val})

        return inputs

    @staticmethod
    def get_outputs(cls) -> list[any]:
        outputs = []
        fields = cls().cls().__dict__
        for key in fields.keys():
            val = fields[key]
            if (
                isinstance(val, Variable)
                and not isinstance(val, InVariable)
                and not isinstance(val, OutVariable)
                and not isinstance(val, OptVariable)
                and val.output
            ):
                outputs.append({"key": key, "val": val})
            elif (
                not isinstance(val, InVariable)
                and not isinstance(val, OutVariable)
                and not isinstance(val, OptVariable)
                and key.lower().startswith("out")
            ):
                outputs.append(
                    {
                        "key": key,
                        "val": Variable(
                            title=Spec.camel_case_to_text(key.lstrip("out")),
                            type=Spec.get_type(val),
                            default=val,
                        ),
                    }
                )

        return outputs

    @staticmethod
    def get_output_vars(cls) -> list[any]:
        outputs = []
        fields = cls().cls().__dict__
        for key in fields.keys():
            val = fields[key]
            if isinstance(val, OutVariable):
                outputs.append({"key": key, "val": val})

        return outputs

    @staticmethod
    def get_options(cls) -> list[any]:
        options = []
        fields = cls().cls().__dict__
        for key in fields.keys():
            val = fields[key]
            if (
                isinstance(val, Variable)
                and not isinstance(val, InVariable)
                and not isinstance(val, OutVariable)
                and not isinstance(val, OptVariable)
                and val.option
            ):
                options.append({"key": key, "val": val})
            elif (
                not isinstance(val, InVariable)
                and not isinstance(val, OutVariable)
                and not isinstance(val, OptVariable)
                and not isinstance(val, Credentials)
                and key.lower().startswith("opt")
            ):
                options.append(
                    {
                        "key": key,
                        "val": Variable(
                            title=Spec.camel_case_to_text(key.lstrip("opt")),
                            type=Spec.get_type(val),
                            default=val,
                        ),
                    }
                )

        return options

    @staticmethod
    def get_credentials(cls) -> list[any]:
        credentials = []
        fields = cls().cls().__dict__
        for key in fields.keys():
            val = fields[key]
            if isinstance(val, Credentials):
                credentials.append({"key": key, "val": val})

        return credentials

    @staticmethod
    def get_option_vars(cls) -> list[any]:
        options = []
        fields = cls().cls().__dict__
        for key in fields.keys():
            val = fields[key]
            if isinstance(val, OptVariable):
                options.append({"key": key, "val": val})

        return options

    @staticmethod
    def camel_case_to_text(str):
        words = [[str[0].upper()]]

        for c in str[1:]:
            if words[-1][-1].islower() and c.isupper():
                words.append(list(c))
            else:
                words[-1].append(c)

        word_arr = ["".join(word) for word in words]
        return " ".join(word_arr)

    @staticmethod
    def get_type(val):
        if isinstance(val, bool):
            return "boolean"
        elif isinstance(val, int) or isinstance(val, float):
            return "number"
        return "string"

    @staticmethod
    def to_snake_case(s):
        return "_".join(
            sub(
                "([A-Z][a-z]+)", r" \1", sub("([A-Z]+)", r" \1", s.replace("-", " "))
            ).split()
        ).lower()



import grpc
import gzip
import json
import enum
import threading
from google.protobuf import json_format
from robomotion import plugin_pb2
from robomotion.plugin_pb2_grpc import RuntimeHelperStub
from google.protobuf.struct_pb2 import Struct
from robomotion.message import Context
from robomotion.error import RuntimeNotInitializedError
from robomotion.capabilities import IsLMOCapable
from robomotion.lmo import IsLMO, DeserializeFromDict, SerializeLMO
from types import SimpleNamespace as Namespace
import base64


class Runtime:
    active_nodes = 0
    client: RuntimeHelperStub = None
    event: threading.Event = threading.Event()
    factories = {}
    nodes = {}
    robotInfo = {}
    @staticmethod
    def set_client(client: RuntimeHelperStub):
        Runtime.client = client

    @staticmethod
    def check_runner_conn(connection: grpc.Channel):
        def cb(state: grpc.ChannelConnectivity):
            if (
                state == grpc.ChannelConnectivity.TRANSIENT_FAILURE
                or state == grpc.ChannelConnectivity.SHUTDOWN
            ):
                Runtime.event.set()

        connection.subscribe(cb, True)

    @staticmethod
    def create_node(name: str, factory):
        Runtime.factories[name] = factory

    @staticmethod
    def add_node(guid: str, node):
        Runtime.nodes[guid] = node

    @staticmethod
    def compress(data: bytes):
        return gzip.compress(data)

    @staticmethod
    def decompress(data: bytes):
        return gzip.decompress(data)

    @staticmethod
    def deserialize(data: bytes, c):
        node = c()
        obj = json.loads(data)
        for key in obj.keys():
            # Special handling for 'func' (LLM Agent instructions)
            if key == 'func' and hasattr(node, 'inFunc'):
                # Overwrite the name attribute of inFunc directly
                node.inFunc._Variable__name = obj['func']
                continue
            if key in node.__dict__.keys():
                if type(obj[key]) is dict:
                    node.__dict__[key] = type(node.__dict__[key])(**obj[key])
                else:
                    node.__dict__[key] = obj[key]

        return node

    @staticmethod
    def close():
        if Runtime.client is None:
            raise RuntimeNotInitializedError

        request = plugin_pb2.Empty()
        Runtime.client.Close(request)

    @staticmethod
    def get_variable(variable, ctx: Context):
        scope = variable.scope
        name = variable.name

        if scope == "Custom":
            return name

        if scope == "Message":
            return ctx.get(name)

        if Runtime.client is None:
            raise RuntimeNotInitializedError

        var = plugin_pb2.Variable(scope=scope, name=name)
        request = plugin_pb2.GetVariableRequest(variable=var)
        response = Runtime.client.GetVariable(request)

        result = json_format.MessageToDict(response.value)["value"]

        if  IsLMOCapable() and IsLMO(result):
            res  = DeserializeFromDict(result)
            
            if res:                 
                result = res.Value()

        return result

    @staticmethod
    def set_variable(variable, ctx: Context, value: object):
        scope = variable.scope
        name = variable.name

        if scope == "Message":
            ctx.set(name, value)
            return

        if Runtime.client is None:
            raise RuntimeNotInitializedError

        val = Struct()
        if IsLMOCapable():
            result = SerializeLMO(value)
            if result:                
                value = result
        val.update({"value": value})


        var = plugin_pb2.Variable(scope=scope, name=name)
        request = plugin_pb2.SetVariableRequest(variable=var, value=val)
        Runtime.client.SetVariable(request)

    @staticmethod
    def get_vault_item(ctx: Context, cred):
        return cred.get_vault_item(ctx=ctx)

    @staticmethod
    def set_vault_item(ctx: Context, cred, data: bytes):
        return cred.set_vault_item(ctx=ctx, data=data)

    class _DefVal:
        def __init__(self, default: object):
            self.default = default

        def __init__(self, scope: str, name: str):
            self.default = {scope: scope, name: name}

    class _Enum:
        def __init__(self, enums: list[str] = [], enumNames: list[str] = []):
            self.__enums = enums
            self.__enumNames = enumNames

        @property
        def enums(self):
            return self.__enums

        @property
        def enumNames(self):
            return self.__enumNames

    @staticmethod
    def get_robot_info() -> dict[str, any]:
        if Runtime.client == None:
            raise RuntimeNotInitializedError
        if len(Runtime.robotInfo) == 0:
            request = plugin_pb2.Empty()
            response = Runtime.client.GetRobotInfo(request)
            Runtime.robotInfo = json_format.MessageToDict(response.robot)["value"]
        return Runtime.robotInfo

    @staticmethod
    def get_robot_version() -> str:
        info = Runtime.get_robot_info()
        return str(info.get("version"))

    @staticmethod
    def download_file(url: str, path: str):
        if Runtime.client == None:
            raise RuntimeNotInitializedError

        request = plugin_pb2.DownloadFileRequest(url=url, path=path)
        Runtime.client.DownloadFile(request)

    @staticmethod
    def app_download(id: str, directory: str, file: str) -> str:
        if Runtime.client == None:
            raise RuntimeNotInitializedError

        request = plugin_pb2.AppDownloadRequest(
            directory=directory, file=file, id=id)

        response = Runtime.client.AppDownload(request)
        return response.path

    @staticmethod
    def app_upload(id: str, path: str) -> str:
        if Runtime.client == None:
            raise RuntimeNotInitializedError

        request = plugin_pb2.AppUploadRequest(id=id, path=path)

        response = Runtime.client.AppUpload(request)
        return response.url

    @staticmethod
    def gateway_request(method: str, endpoint: str, body: str, headers: map) -> plugin_pb2.GatewayRequestResponse:
        if Runtime.client == None:
            raise RuntimeNotInitializedError

        request = plugin_pb2.GatewayRequestRequest(
            method=method, endpoint=endpoint, body=body, headers=headers
        )

        return Runtime.client.GatewayRequest(request)

    @staticmethod
    def is_running() -> bool:
        """Check if the runtime is running via gRPC call."""
        if Runtime.client is None:
            raise RuntimeNotInitializedError
        request = plugin_pb2.Empty()
        response = Runtime.client.IsRunning(request)
        return response.isRunning

    @staticmethod
    def get_port_connections(guid: str, port: int) -> list[dict]:
        """Get information about nodes connected to a specific port via gRPC call."""
        if Runtime.client is None:
            raise RuntimeNotInitializedError
        request = plugin_pb2.GetPortConnectionsRequest(guid=guid, port=port)
        response = Runtime.client.GetPortConnections(request)
        result = []
        for node in response.nodes:
            # node.config is bytes, decode as JSON after double base64 decode
            try:
                config_dict = json.loads(node.config)
                # Double base64 decode on the 'config' field if present
                if 'config' in config_dict:
                    decoded = base64.b64decode(config_dict['config'])
                    decoded = base64.b64decode(decoded)
                    config_dict = json.loads(decoded)
            except Exception:
                config_dict = {}
            result.append({
                'type': node.type,
                'version': node.version,
                'config': config_dict
            })
        return result



# port.py
# Python equivalent of Go's port.go

from typing import List, Dict, Any

# Port is a type alias for a list of GUIDs of connected nodes
class Port(list):
    def __init__(self, direction, position, name, icon=None, color=None, filters=None, order=None):
        super().__init__()
        self.direction = direction
        self.position = position
        self.name = name
        self.icon = icon
        self.color = color
        self.filters = filters if filters is not None else []
        self.order = order



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: runner.proto
# Protobuf Python Version: 5.27.2
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    27,
    2,
    '',
    'runner.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0crunner.proto\x12\x05proto\"\x06\n\x04Null\"\x82\x01\n\x11InitRunnerRequest\x12\x0e\n\x06\x61piUrl\x18\x01 \x01(\t\x12\r\n\x05token\x18\x02 \x01(\t\x12\x14\n\x0crobot_server\x18\x03 \x01(\r\x12\x10\n\x08natsPort\x18\x04 \x01(\r\x12\x12\n\nrobot_name\x18\x05 \x01(\t\x12\x12\n\nsend_crash\x18\x06 \x01(\x08\"T\n\nRunRequest\x12\x0e\n\x06repoid\x18\x01 \x01(\t\x12\x12\n\nrepository\x18\x02 \x01(\t\x12\x11\n\tnamespace\x18\x03 \x01(\t\x12\x0f\n\x07version\x18\x04 \x01(\t\"\'\n\x11RobotNameResponse\x12\x12\n\nrobot_name\x18\x01 \x01(\t\"\x05\n\x03Nil\"\x1f\n\rAttachRequest\x12\x0e\n\x06\x63onfig\x18\x01 \x01(\x0c\"\"\n\rDetachRequest\x12\x11\n\tnamespace\x18\x01 \x01(\t2\xb5\x01\n\x06Runner\x12-\n\x04Init\x12\x18.proto.InitRunnerRequest\x1a\x0b.proto.Null\x12%\n\x03Run\x12\x11.proto.RunRequest\x1a\x0b.proto.Null\x12!\n\x05\x43lear\x12\x0b.proto.Null\x1a\x0b.proto.Null\x12\x32\n\tRobotName\x12\x0b.proto.Null\x1a\x18.proto.RobotNameResponse2\r\n\x0bRobotHelper2_\n\x05\x44\x65\x62ug\x12*\n\x06\x41ttach\x12\x14.proto.AttachRequest\x1a\n.proto.Nil\x12*\n\x06\x44\x65tach\x12\x14.proto.DetachRequest\x1a\n.proto.NilB*Z(robomotion/robomotion-runner/robot/protob\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'runner_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'Z(robomotion/robomotion-runner/robot/proto'
  _globals['_NULL']._serialized_start=23
  _globals['_NULL']._serialized_end=29
  _globals['_INITRUNNERREQUEST']._serialized_start=32
  _globals['_INITRUNNERREQUEST']._serialized_end=162
  _globals['_RUNREQUEST']._serialized_start=164
  _globals['_RUNREQUEST']._serialized_end=248
  _globals['_ROBOTNAMERESPONSE']._serialized_start=250
  _globals['_ROBOTNAMERESPONSE']._serialized_end=289
  _globals['_NIL']._serialized_start=291
  _globals['_NIL']._serialized_end=296
  _globals['_ATTACHREQUEST']._serialized_start=298
  _globals['_ATTACHREQUEST']._serialized_end=329
  _globals['_DETACHREQUEST']._serialized_start=331
  _globals['_DETACHREQUEST']._serialized_end=365
  _globals['_RUNNER']._serialized_start=368
  _globals['_RUNNER']._serialized_end=549
  _globals['_ROBOTHELPER']._serialized_start=551
  _globals['_ROBOTHELPER']._serialized_end=564
  _globals['_DEBUG']._serialized_start=566
  _globals['_DEBUG']._serialized_end=661
# @@protoc_insertion_point(module_scope)
